import rclpy
from rclpy.node import Node
from rclpy.action import ActionClient, ActionServer
from rclpy.duration import Duration

from franka_msgs.action import Grasp
from franka_msgs.action import Move
from control_msgs.action import FollowJointTrajectory
from trajectory_msgs.msg import JointTrajectory, JointTrajectoryPoint
from franka_cps_msgs.action import GraspSequence

from threading import Event

import sqlite3

class GraspDemoNode(Node):
    def __init__(self):
        super().__init__('grasp_demo_node')

        # Event Handling
        self.action_completed_event = Event()
        self._action_result = None
        # Grasp Sequence Service
        self.execute_action = ActionServer(self, GraspSequence,
                                           'franka_grasp/grasp_sequence', self.execute_callback)

        # Franka Position Controller
        self.position_client = ActionClient(self, FollowJointTrajectory,
                                            'panda_arm_controller/follow_joint_trajectory')

        self._send_position_future = None
        self._position_result_future = None

        # Franka Gripper Action Server clients for Move and Grasp
        self.franka_move_action_client = ActionClient(self, Move, 'panda_gripper/move')
        self.franka_grasp_action_client = ActionClient(self, Grasp, 'panda_gripper/grasp')

        self._send_move_future = None
        self._move_result_future = None
        self._send_grasp_future = None
        self._grasp_result_future = None

        # Sqlite Database
        self.conn = None
        self.cursor = None
        self.database_name = '/home/niko/sqlite_dbs/logistics_db'

        self.init_db()

    def init_db(self):
        # Connect to the sqlite db
        self.conn = sqlite3.connect(self.database_name)
        self.cursor = self.conn.cursor()

        self.cursor.execute(
            '''CREATE TABLE IF NOT EXISTS position_table (
            position_id INTEGER PRIMARY KEY AUTOINCREMENT,
            position_name TEXT UNIQUE NOT NULL,
            joint_1 REAL NOT NULL, 
            joint_2 REAL NOT NULL, 
            joint_3 REAL NOT NULL, 
            joint_4 REAL NOT NULL, 
            joint_5 REAL NOT NULL,
            joint_6 REAL NOT NULL,
            joint_7 REAL NOT NULL)''')

        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS position_sequence_table (
                sequence_id INTEGER PRIMARY KEY AUTOINCREMENT,
                sequence_name TEXT UNIQUE NOT NULL,
                position_idle TEXT NOT NULL,
                position_hower TEXT NOT NULL,
                position_grasp TEXT NOT NULL,
                position_drop TEXT NOT NULL,
                FOREIGN KEY (position_idle) REFERENCES position_table (position_name),
                FOREIGN KEY (position_hower) REFERENCES position_table (position_name),
                FOREIGN KEY (position_grasp) REFERENCES position_table (position_name),
                FOREIGN KEY (position_drop) REFERENCES position_table (position_name)
            )
        ''')
        self.conn.commit()
        self.get_logger().info('Initializing Grasp Demo Node finished!')

    def execute_callback(self, goal_handle):
        self.get_logger().info(' Executing Grasp Sequence')
        feedback = GraspSequence.Feedback()

        sequence = [
            (self.arm_move_to_position, (goal_handle.request.hower, 10), "Moving to Hower Position"),
            (self.arm_move_to_position, (goal_handle.request.grasp, 10), "Moving to Grasp Position"),
            (self.gripper_grasp, None, "Grasping Object"),
            (self.arm_move_to_position, (goal_handle.request.hower, 10), "Moving back to Hower Position"),
            (self.arm_move_to_position, (goal_handle.request.drop, 10), "Moving back to Drop Position"),
            (self.gripper_move, 0.08, "Opening Gripper"),
            (self.arm_move_to_position, (goal_handle.request.idle, 5), "Moving to Idle Position")
        ]

        for action, args, step_descrioption in sequence:
            feedback.current_step = step_descrioption
            goal_handle.publish_feedback(feedback)
            # if goal_handle.is_cancle_requested():
            #     goal_handle.canceled()
            #     return GraspSequence.Result(success=False, error_code="Canceled")

            success = action(*args) if args else action()
            if not success:
                goal_handle.abort()
                return GraspSequence.Result(success=False, error_code="aborted")

        goal_handle.succeed()
        return GraspSequence.Result(sucess=True)


    # region Position Controller
    def arm_move_to_position(self, joint_positions, duration):
        self.action_completed_event.clear()  # Reset the event before sending a new goal

        goal_msg = FollowJointTrajectory.Goal()
        trajectory = JointTrajectory()
        trajectory.joint_names = ['panda_joint1', 'panda_joint2',
                                  'panda_joint3', 'panda_joint4',
                                  'panda_joint5', 'panda_joint6',
                                  'panda_joint7']
        point = JointTrajectoryPoint()
        point.positions = joint_positions.position
        time = Duration(seconds=duration)
        self.get_logger().error(f"Time: {time}")
        point.time_from_start = time.to_msg()
        self.get_logger().error(f"NEW Point{point}")
        trajectory.points.append(point)
        goal_msg.trajectory = trajectory
        self.get_logger().error(f'Goal:{goal_msg}')

        self.position_client.wait_for_server()
        self._send_position_future = self.position_client.send_goal_async(goal_msg)
        self._send_position_future.add_done_callback(self.position_response_callback)
        # Wait here until the action completes
        self.action_completed_event.wait()

        # After the event is set in position_result_callback
        if hasattr(self, '_action_result') and self._action_result:
            self._action_result = None
            return True
        return False

    def position_response_callback(self, future):
        goal_handle = future.result()
        if not goal_handle.accepted:
            self.get_logger().info('Positon goal rejected!')
            self.action_completed_event.set()  # Signal that the action is completed (with failure)
            return

        self.get_logger().info('Position goal accepted!')
        self._position_result_future = goal_handle.get_result_async()
        self._position_result_future.add_done_callback(self.position_result_callback)

    def position_result_callback(self, future):
        result = future.result().result
        self.get_logger().info(f'Position result: Success: {result.success}, Error: {result.error}')
        self._action_result = result.success  # Store the result for checking after wait
        self.action_completed_event.set()     # Signal that the action is completed
    # endregion

    # region Grasp Action
    def gripper_grasp(self):
        self.action_completed_event.clear()  # Reset the event before sending a new goal

        goal_msg = Grasp()
        goal_msg.width = 0.0
        goal_msg.speed = 0.1
        goal_msg.force = 50.0

        self.franka_grasp_action_client.wait_for_server()
        self._send_grasp_future = self.franka_grasp_action_client.send_goal_async(goal_msg, feedback_callback=self.action_fb_callback)
        self._send_grasp_future.add_done_callback(self.grasp_response_callback)
        self.action_completed_event.wait()

        # After the event is set in grasp_result_callback
        if hasattr(self, '_action_result') and self._action_result:
            self._action_result = None
            return True
        return False

    def action_fb_callback(self, msg):
        feedback = msg.feedback
        self.get_logger().info(f'Received feedback: Current Width: {feedback.current_width}')

    def grasp_response_callback(self, future):
        goal_handle = future.result()
        if not goal_handle.accepted:
            self.get_logger().info('Grasp goal rejected!')
            self.action_completed_event.set()  # Signal that the action is completed (with failure)
            return

        self.get_logger().info('Grasp goal accepted!')
        self._grasp_result_future = goal_handle.get_result_async()
        self._grasp_result_future.add_done_callback(self.grasp_result_callback)

    def grasp_result_callback(self, future):
        result = future.result().result
        self.get_logger().info(f'Grasp result: Success: {result.success}, Error: {result.error}')
        self._action_result = result.success  # Store the result for checking after wait
        self.action_completed_event.set()     # Signal that the action is completed
    # endregion

    # region Move Action
    def gripper_move(self, width):
        goal_msg = Move()
        goal_msg.width = width
        goal_msg.speed = 0.1

        self.franka_move_action_client.wait_for_server()
        self._send_move_future = self.franka_move_action_client.send_goal_async(goal_msg, feedback_callback=self.action_fb_callback)
        self._send_move_future.add_done_callback(self.move_response_callback)
        self.action_completed_event.wait()

        # After the event is set in move_result_callback
        if hasattr(self, '_action_result') and self._action_result:
            self._action_result = None
            return True
        return False

    def move_response_callback(self, future):
        goal_handle = future.result()
        if not goal_handle.accepted:
            self.get_logger().info('Move goal rejected!')
            self.action_completed_event.set()  # Signal that the action is completed (with failure)
            return

        self.get_logger().info('Move goal accepted!')
        self._move_result_future = goal_handle.get_result_async()
        self._move_result_future.add_done_callback(self.move_result_callback)

    def move_result_callback(self, future):
        result = future.result().result
        self.get_logger().info(f'Move result: Success: {result.success}, Error: {result.error}')
        self._action_result = result.success  # Store the result for checking after wait
        self.action_completed_event.set()     # Signal that the action is completed
    # endregion


def main(args=None):
    rclpy.init(args=args)
    node = GraspDemoNode()
    rclpy.spin(node)
    rclpy.shutdown()

if __name__ == '__main__':
    main()