import rclpy
from rclpy.node import Node

from geometry_msgs.msg import PoseStamped
from std_msgs.msg import String

import sqlite3


class PoseLearnNode(Node):
    def __init__(self):
        super().__init__('pose_learn_node')

        self.current_pose = PoseStamped()

        self.pose_sub = self.create_subscription(PoseStamped, 'moveit_interface/current_pose',
                                                 self.pose_callback, 10)

        self.save_sub = self.create_subscription(String, '/logistic/save_pose', self.save_pose_callback, 10)

        self.seq_sub = self.create_subscription(String, '/logistic/save_seq', self.save_sequence_callback, 10)

        # Connect to Database
        self.conn = None
        self.cursor = None
        self.database_name = '/home/niko/sqlite_dbs/logistics_db'

        self.init_db()

    def init_db(self):
        # Connect to the pose db
        self.conn = sqlite3.connect(self.database_name)
        self.conn.execute("PRAGMA foreign_keys = ON")
        self.cursor = self.conn.cursor()
        self.cursor.execute(
            '''CREATE TABLE IF NOT EXISTS pose_table (
            pose_id INTEGER PRIMARY KEY AUTOINCREMENT,
            pose_name TEXT UNIQUE NOT NULL,
            position_x REAL NOT NULL, 
            position_y REAL NOT NULL, 
            position_z REAL NOT NULL,
            orientation_x REAL NOT NULL, 
            orientation_y REAL NOT NULL, 
            orientation_z REAL NOT NULL, 
            orientation_w REAL NOT NULL)''')

        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS sequence_table (
                sequence_id INTEGER PRIMARY KEY AUTOINCREMENT,
                sequence_name TEXT UNIQUE NOT NULL,
                pose_idle TEXT NOT NULL,
                pose_hower TEXT NOT NULL,
                pose_grasp TEXT NOT NULL,
                pose_drop TEXT NOT NULL,
                FOREIGN KEY (pose_idle) REFERENCES pose_table (pose_name),
                FOREIGN KEY (pose_hower) REFERENCES pose_table (pose_name),
                FOREIGN KEY (pose_grasp) REFERENCES pose_table (pose_name),
                FOREIGN KEY (pose_drop) REFERENCES pose_table (pose_name)
            )
        ''')
        self.conn.commit()
        self.get_logger().info('Initializing PoseLearnNode')

    def pose_callback(self, msg):
        self.current_pose = msg

    def save_pose_callback(self, msg):
        pose_name = msg.data
        position = self.current_pose.pose.position
        position = self.current_pose.pose.position
        orientation = self.current_pose.pose.orientation
        try:

            self.cursor.execute('''INSERT INTO pose_table
                                (pose_name, position_x, position_y, position_z,
                                 orientation_x, orientation_y, orientation_z, orientation_w)
                                VALUES (?,?, ?, ?, ?, ?, ?, ?)''',
                                (pose_name, position.x, position.y, position.z,
                                 orientation.x, orientation.y, orientation.z, orientation.w))
            self.conn.commit()
        except sqlite3.IntegrityError:
            self.get_logger().info(f"Pose with name '{pose_name}' already exists.")

    def save_sequence_callback(self, msg):
        try:
            self.cursor.execute('''INSERT INTO sequence_table
            (sequence_name, pose_idle, pose_hower, pose_grasp, pose_drop)
             VALUES (?, ?, ?, ?, ?)''',
                                (msg.sequence_name, msg.pose_idle, msg.pose_hower, msg.pose_grasp, msg.pose_drop))
            self.conn.commit()
        except sqlite3.IntegrityError:
            self.get_logger().info(f"Sequence with name '{msg.sequence_name}' already exists.")


def main(args=None):
    rclpy.init(args=args)
    pose_learn_node = PoseLearnNode()
    rclpy.spin(pose_learn_node)
    pose_learn_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
