import threading
import tkinter as tk
from tkinter import ttk

import sqlite3

import rclpy
from rclpy.node import Node

from sensor_msgs.msg import JointState


class GUINode(Node):
    def __init__(self):
        super().__init__('position_learn_gui_node')

        self.create_subscription(JointState, 'franka/joint_states', self.position_callback, 10)

        self.current_position = {
            'joint_1': 0.0,
            'joint_2': 0.0,
            'joint_3': 0.0,
            'joint_4': 0.0,
            'joint_5': 0.0,
            'joint_6': 0.0,
            'joint_7': 0.0
        }

    def position_callback(self, msg):
        self.current_position['joint_1'] = msg.position[0]
        self.current_position['joint_2'] = msg.position[1]
        self.current_position['joint_3'] = msg.position[2]
        self.current_position['joint_4'] = msg.position[3]
        self.current_position['joint_5'] = msg.position[4]
        self.current_position['joint_6'] = msg.position[5]
        self.current_position['joint_7'] = msg.position[6]


class MyGUI:
    def __init__(self, root, node):
        self.root = root
        self.root.title("Franka Grasp Interface")

        # ROS2 Node
        self.node = node

        # Connect to your SQLite database
        self.conn = sqlite3.connect('/home/niko/sqlite_dbs/logistics_db')
        self.cursor = self.conn.cursor()

        # Track the current mode
        self.current_mode = 'Learn'  # Starts in Learn mode

        # Initialize frames for Learn and Run modes
        self.learn_frame = ttk.Frame(root)

        # Position Submenu attributes
        self.position_name = ''
        self.position_entries = {}
        self.position_entries_values = {
            'joint_1': 0.0,
            'joint_2': 0.0,
            'joint_3': 0.0,
            'joint_4': 0.0,
            'joint_5': 0.0,
            'joint_6': 0.0,
            'joint_7': 0.0
        }

        # Grasp Submenu attributes
        self.grasp_name = ''
        self.idle = ''
        self.hower = ''
        self.grasp = ''
        self.drop = ''

        # Database Submenu attributes
        self.position_db = ''
        self.grasp_db = ''

        # Grasp Run Submenu attributes
        self.grasp_run = ''
        self.grasp_entries = {}

        # GUI Subcomponents
        self.position_name_entry = None

        self.grasp_name_entry = None
        self.idle_selector = None
        self.hower_selector = None
        self.grasp_selector = None
        self.drop_selector = None
        self.width_grasp_entry = None
        self.width_idle_entry = None

        self.position_db_selector = None
        self.grasp_db_selector = None

        # Frames
        self.position_frame = None
        self.grasp_frame = None
        self.db_frame = None
        self.separator_1 = None
        self.separator_2 = None

        # Initially set up Learn mode
        self.setup_learn_mode()

    # <editor-fold desc="Menu Methods">

    def setup_learn_mode(self):
        self.create_position_column()
        self.separator_1 = tk.Canvas(self.root, width=1, bg="grey")
        self.separator_1.pack(side=tk.LEFT, fill=tk.Y)
        self.create_grasp_column()
        self.separator_2 = tk.Canvas(self.root, width=1, bg="grey")
        self.separator_2.pack(side=tk.LEFT, fill=tk.Y)
        self.create_db_edit_column()

    # </editor-fold>

    # <editor-fold desc="Initialize Methods">
    def set_initial_position_values(self):
        initial_values = {
            'joint_1': 0.0,
            'joint_2': 0.0,
            'joint_3': 0.0,
            'joint_4': 0.0,
            'joint_5': 0.0,
            'joint_6': 0.0,
            'joint_7': 0.0
        }
        for key, value in initial_values.items():
            if key in self.position_entries:
                self.position_entries[key].config(state='normal')  # Enable writing
                self.position_entries[key].delete(0, tk.END)
                self.position_entries[key].insert(0, str(value))
                self.position_entries[key].config(state='readonly')  # Set back to read-only

    # </editor-fold>

    # <editor-fold desc="SQLite Methods">
    def get_position_names(self):
        self.cursor.execute("SELECT position_name FROM position_table")
        return [row[0] for row in self.cursor.fetchall()]

    def get_grasp_names(self):
        self.cursor.execute("SELECT sequence_name FROM position_sequence_table")
        return [row[0] for row in self.cursor.fetchall()]

    def get_position_by_name(self, position_name):
        self.cursor.execute(
            "SELECT joint_1, joint_2, joint_3, joint_4, joint_5, joint_6, joint_7 FROM position_table WHERE position_name = ?",
            (position_name,))
        result = self.cursor.fetchone()
        if result:
            position = JointState()
            position.position = result
            return position
        else:
            print("Position not found")
            return None

    def get_grasp_sequence_by_name(self, sequence_name):
        self.cursor.execute(
            "SELECT position_idle, position_hower, position_grasp, position_drop"
            " FROM position_sequence_table WHERE sequence_name = ?",
            (sequence_name,))
        result = self.cursor.fetchone()
        if result:
            self.grasp_entries['idle_position'] = result[0]
            self.grasp_entries['hower_position'] = result[1]
            self.grasp_entries['grasp_position'] = result[2]
            self.grasp_entries['drop_position'] = result[3]
        else:
            print("Sequence not found")

    def add_new_position(self, position_name, joint_1, joint_2, joint_3, joint_4, joint_5, joint_6, joint_7):
        self.cursor.execute(
            "INSERT INTO position_table (position_name, joint_1, joint_2, joint_3, joint_4, joint_5, joint_6, joint_7)"
            " VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
            (position_name,
             joint_1,
             joint_2,
             joint_3,
             joint_4,
             joint_5,
             joint_6,
             joint_7))

        self.conn.commit()

    def add_new_grasp(self, sequence_name, idle, hower, grasp, drop):
        print(sequence_name, idle, hower, grasp, drop)
        self.cursor.execute("INSERT INTO position_sequence_table"
                            """ (sequence_name, position_idle, position_hower, position_grasp, position_drop)
                             VALUES (?, ?, ?, ?, ?)""",
                            (sequence_name, idle, hower, grasp, drop))

        self.conn.commit()

    def delete_position_by_name(self, position_name):
        self.cursor.execute("DELETE FROM position_table WHERE position_name = ?", (position_name,))

        self.conn.commit()

    def delete_grasp_by_name(self, sequence_name):
        self.cursor.execute("DELETE FROM position_sequence_table WHERE sequence_name = ?", (sequence_name,))

        self.conn.commit()

    # </editor-fold>

    # <editor-fold desc="Update Methods">
    def update_position_values(self):
        for key, value in self.node.current_position.items():
            if key in self.position_entries:
                self.position_entries[key].config(state='normal')  # Enable writing
                self.position_entries[key].delete(0, tk.END)
                self.position_entries[key].insert(0, f"{value:.5f}")
                self.position_entries[key].config(state='readonly')  # Set back to read-only

    def update_position_selectors(self, db_positions):
        self.idle_selector['values'] = db_positions
        self.hower_selector['values'] = db_positions
        self.grasp_selector['values'] = db_positions
        self.drop_selector['values'] = db_positions

    def update_db_selectors(self, db_positions, db_grasp):
        self.position_db_selector['values'] = db_positions
        self.grasp_db_selector['values'] = db_grasp

    def update_from_database(self):
        db_positions = self.get_position_names()
        db_grasps = self.get_grasp_names()
        self.update_position_selectors(db_positions)
        self.update_db_selectors(db_positions, db_grasps)

    # </editor-fold>

    # <editor-fold desc="SubMenu Events">
    # Position Events
    def position_set_event(self):
        self.position_name = self.position_name_entry.get()

        self.add_new_position(self.position_name,
                          float(self.position_entries['joint_1'].get()),
                          float(self.position_entries['joint_2'].get()),
                          float(self.position_entries['joint_3'].get()),
                          float(self.position_entries['joint_4'].get()),
                          float(self.position_entries['joint_5'].get()),
                          float(self.position_entries['joint_6'].get()),
                          float(self.position_entries['joint_7'].get()))
        print("Position Button Clicked! Position Name:", self.position_name)

    # Grasp Events
    def idle_selector_event(self, event):
        self.idle = self.idle_selector.get()

    def hower_selector_event(self, event):
        self.hower = self.hower_selector.get()

    def grasp_selector_event(self, event):
        self.grasp = self.grasp_selector.get()

    def drop_selector_event(self, event):
        self.drop = self.drop_selector.get()

    def grasp_set_event(self):
        self.grasp_name = self.grasp_name_entry.get()
        self.add_new_grasp(self.grasp_name,
                           self.idle,
                           self.hower,
                           self.grasp,
                           self.drop)
        print("Grasp Button Clicked! Grasp Name:", self.grasp_name)

    # Database Events
    def position_db_selector_event(self, event):
        self.position_db = self.position_db_selector.get()

    def grasp_db_selector_event(self, event):
        self.grasp_db = self.grasp_db_selector.get()

    def position_delete_event(self):
        self.delete_position_by_name(self.position_db)
        print("Position: ", self.position_db, ' was deleted!')

    def grasp_delete_event(self):
        self.delete_grasp_by_name(self.grasp_db)
        print("Grasp: ", self.grasp_db, ' was deleted!')

    # Run Events
    def grasp_run_selector_event(self, event):
        self.grasp_run = self.grasp_run_selector.get()
        self.get_grasp_sequence_by_name(self.grasp_run)

    def grasp_perform_event(self):
        idle = self.get_position_by_name(self.idle)
        hower = self.get_position_by_name(self.hower)
        grasp = self.get_position_by_name(self.grasp)
        drop = self.get_position_by_name(self.drop)

        feedback = self.node.send_grasp_sequence_goal(idle, hower, grasp, drop)

        print("Grasp: ", self.grasp_run, ' is performed!')

    # </editor-fold>

    # <editor-fold desc="Validate methods">
    def validate_float(self, P):
        try:
            value = float(P)
            if self.width_lower <= value <= self.width_upper:
                return True
            else:
                return False
        except ValueError:
            return False

    # </editor-fold>

    # <editor-fold desc="SubMenu Create Methods">
    def create_position_column(self):
        self.position_frame = tk.Frame(self.root)
        self.position_frame.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

        submenu_label = tk.Label(self.position_frame, text="New Position", font=("Arial", 12, "bold"))
        submenu_label.grid(row=0, column=0, columnspan=2)

        # Name Entry
        tk.Label(self.position_frame, text="Position Name").grid(row=1, column=0)
        self.position_name_entry = tk.Entry(self.position_frame)
        self.position_name_entry.grid(row=1, column=1)

        # Separator
        separator = ttk.Separator(self.position_frame, orient='horizontal')
        separator.grid(row=2, column=0, columnspan=2, sticky='ew')

        # Title for Current Position
        current_position_label = tk.Label(self.position_frame, text="Current Position", font=("Arial", 10, "bold"))
        current_position_label.grid(row=3, column=0, columnspan=2)

        # Separator
        separator = ttk.Separator(self.position_frame, orient='horizontal')
        separator.grid(row=4, column=0, columnspan=2, sticky='ew')

        # Position Values Entries
        position_labels = ['Joint 1', 'Joint 2', 'Joint 3', 'Joint 4', 'Joint 5', 'Joint 6', 'Joint 7']
        for i, label in enumerate(position_labels, start=5):
            tk.Label(self.position_frame, text=label).grid(row=i, column=0)
            entry = tk.Entry(self.position_frame, state='readonly', width=10)
            entry.grid(row=i, column=1)
            self.position_entries[label.replace(' ', '_').lower()] = entry

        self.set_initial_position_values()

        # Separator
        separator = ttk.Separator(self.position_frame, orient='horizontal')
        separator.grid(row=14, column=0, columnspan=2, sticky='ew')

        # Set Button
        set_button = tk.Button(self.position_frame, text="Set New Position", command=self.position_set_event)
        set_button.grid(row=15, column=0, columnspan=2)

        self.position_frame.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

    def create_grasp_column(self):
        self.grasp_frame = tk.Frame(self.root)
        self.grasp_frame.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

        submenu_label = tk.Label(self.grasp_frame, text="New Grasp", font=("Arial", 12, "bold"))
        submenu_label.grid(row=0, column=0, columnspan=2)

        # Name Entry
        tk.Label(self.grasp_frame, text="Grasp Name").grid(row=1, column=0)
        self.grasp_name_entry = tk.Entry(self.grasp_frame)
        self.grasp_name_entry.grid(row=1, column=1)

        # Separator
        separator = ttk.Separator(self.grasp_frame, orient='horizontal')
        separator.grid(row=2, column=0, columnspan=2, sticky='ew')

        # Title for Current Position
        current_position_label = tk.Label(self.grasp_frame, text="Grasp Position", font=("Arial", 10, "bold"))
        current_position_label.grid(row=3, column=0, columnspan=2)

        # Separator
        separator = ttk.Separator(self.grasp_frame, orient='horizontal')
        separator.grid(row=4, column=0, columnspan=2, sticky='ew')

        # Idle Position Selector
        idle_label = tk.Label(self.grasp_frame, text="Idle Position", font="Arial")
        idle_label.grid(row=5, column=0)
        self.idle_selector = ttk.Combobox(self.grasp_frame, values=[""], width=10)
        self.idle_selector.grid(row=5, column=1)
        self.idle_selector.bind('<<ComboboxSelected>>', self.idle_selector_event)

        # Hower Position Selector
        hower_label = tk.Label(self.grasp_frame, text="Hower Position", font="Arial")
        hower_label.grid(row=6, column=0)
        self.hower_selector = ttk.Combobox(self.grasp_frame, values=[""], width=10)
        self.hower_selector.grid(row=6, column=1)
        self.hower_selector.bind('<<ComboboxSelected>>', self.hower_selector_event)

        # Grasp Position Selector
        grasp_label = tk.Label(self.grasp_frame, text="Grasp Position", font="Arial")
        grasp_label.grid(row=7, column=0)
        self.grasp_selector = ttk.Combobox(self.grasp_frame, values=[""], width=10)
        self.grasp_selector.grid(row=7, column=1)
        self.grasp_selector.bind('<<ComboboxSelected>>', self.grasp_selector_event)

        # Drop Position Selector
        drop_label = tk.Label(self.grasp_frame, text="Drop Position", font="Arial")
        drop_label.grid(row=8, column=0)
        self.drop_selector = ttk.Combobox(self.grasp_frame, values=[""], width=10)
        self.drop_selector.grid(row=8, column=1)
        self.drop_selector.bind('<<ComboboxSelected>>', self.drop_selector_event)

        # Separator
        separator = ttk.Separator(self.grasp_frame, orient='horizontal')
        separator.grid(row=12, column=0, columnspan=2, sticky='ew')

        # Set Button
        set_button = tk.Button(self.grasp_frame, text="Set New Grasp", command=self.grasp_set_event)
        set_button.grid(row=13, column=0, columnspan=2)

    def create_db_edit_column(self):
        self.db_frame = tk.Frame(self.root)
        self.db_frame.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

        submenu_label = tk.Label(self.db_frame, text="Edit Database", font=("Arial", 12, "bold"))
        submenu_label.grid(row=0, column=0, columnspan=2)

        # Separator
        separator = ttk.Separator(self.db_frame, orient='horizontal')
        separator.grid(row=1, column=0, columnspan=2, sticky='ew')

        # Position Database Selector
        position_db_label = tk.Label(self.db_frame, text="Positions", font="Arial")
        position_db_label.grid(row=2, column=0)
        self.position_db_selector = ttk.Combobox(self.db_frame, values=[""], width=8)
        self.position_db_selector.grid(row=2, column=1)
        self.position_db_selector.bind('<<ComboboxSelected>>', self.position_db_selector_event)

        # Position Delete Button
        position_delete_button = tk.Button(self.db_frame, text="Delete Position", command=self.position_delete_event)
        position_delete_button.grid(row=3, column=0, columnspan=2)

        # Separator
        separator = ttk.Separator(self.db_frame, orient='horizontal')
        separator.grid(row=4, column=0, columnspan=2, sticky='ew')

        # Grasp Database Selector
        grasp_db_label = tk.Label(self.db_frame, text="Grasp", font="Arial")
        grasp_db_label.grid(row=5, column=0)
        self.grasp_db_selector = ttk.Combobox(self.db_frame, values=[""], width=8)
        self.grasp_db_selector.grid(row=5, column=1)
        self.grasp_db_selector.bind('<<ComboboxSelected>>', self.grasp_db_selector_event)

        # Grasp Delete Button
        grasp_delete_button = tk.Button(self.db_frame, text="Delete Grasp", command=self.grasp_delete_event)
        grasp_delete_button.grid(row=6, column=0, columnspan=2)
    # </editor-fold>


def main(args=None):
    rclpy.init(args=args)
    gui_node = GUINode()

    thread_spin = threading.Thread(target=rclpy.spin, args=(gui_node,))
    thread_spin.start()

    root = tk.Tk()

    # app = MyGUI(root)
    app = MyGUI(root, gui_node)
    while True:
        app.update_from_database()
        app.update_position_values()
        root.update()

    gui_node.destroy_node()
    rclpy.shutdown()
    app.conn.close()


if __name__ == '__main__':
    main()
